@@ -1,12 +1,14 @@
-import ast
 import codecs
+import json
 import os
 import pipes
 import re
 import sys
 import warnings
+from itertools import chain
 
 import py
+from pkg_resources import to_filename
 
 import tox
 
@@ -14,10 +16,19 @@ from .config import DepConfig
 
 
 class CreationConfig:
-    def __init__(self, md5, python, version, sitepackages, usedevelop, deps, alwayscopy):
-        self.md5 = md5
-        self.python = python
-        self.version = version
+    def __init__(
+        self,
+        base_resolved_python_md5,
+        base_resolved_python_path,
+        tox_version,
+        sitepackages,
+        usedevelop,
+        deps,
+        alwayscopy,
+    ):
+        self.base_resolved_python_md5 = base_resolved_python_md5
+        self.base_resolved_python_path = base_resolved_python_path
+        self.tox_version = tox_version
         self.sitepackages = sitepackages
         self.usedevelop = usedevelop
         self.alwayscopy = alwayscopy
@@ -25,49 +36,71 @@ class CreationConfig:
 
     def writeconfig(self, path):
         lines = [
-            "{} {}".format(self.md5, self.python),
+            "{} {}".format(self.base_resolved_python_md5, self.base_resolved_python_path),
             "{} {:d} {:d} {:d}".format(
-                self.version, self.sitepackages, self.usedevelop, self.alwayscopy
+                self.tox_version, self.sitepackages, self.usedevelop, self.alwayscopy
             ),
         ]
         for dep in self.deps:
             lines.append("{} {}".format(*dep))
+        content = "\n".join(lines)
         path.ensure()
-        path.write("\n".join(lines))
+        path.write(content)
+        return content
 
     @classmethod
     def readconfig(cls, path):
         try:
             lines = path.readlines(cr=0)
-            value = lines.pop(0).split(None, 1)
-            md5, python = value
-            version, sitepackages, usedevelop, alwayscopy = lines.pop(0).split(None, 4)
+            base_resolved_python_info = lines.pop(0).split(None, 1)
+            tox_version, sitepackages, usedevelop, alwayscopy = lines.pop(0).split(None, 4)
             sitepackages = bool(int(sitepackages))
             usedevelop = bool(int(usedevelop))
             alwayscopy = bool(int(alwayscopy))
             deps = []
             for line in lines:
-                md5, depstring = line.split(None, 1)
-                deps.append((md5, depstring))
-            return CreationConfig(md5, python, version, sitepackages, usedevelop, deps, alwayscopy)
+                base_resolved_python_md5, depstring = line.split(None, 1)
+                deps.append((base_resolved_python_md5, depstring))
+            base_resolved_python_md5, base_resolved_python_path = base_resolved_python_info
+            return CreationConfig(
+                base_resolved_python_md5,
+                base_resolved_python_path,
+                tox_version,
+                sitepackages,
+                usedevelop,
+                deps,
+                alwayscopy,
+            )
         except Exception:
             return None
 
+    def matches_with_reason(self, other, deps_matches_subset=False):
+        for attr in (
+            "base_resolved_python_md5",
+            "base_resolved_python_path",
+            "tox_version",
+            "sitepackages",
+            "usedevelop",
+            "alwayscopy",
+        ):
+            left = getattr(self, attr)
+            right = getattr(other, attr)
+            if left != right:
+                return False, "attr {} {!r}!={!r}".format(attr, left, right)
+        self_deps = set(self.deps)
+        other_deps = set(other.deps)
+        if self_deps != other_deps:
+            if deps_matches_subset:
+                diff = other_deps - self_deps
+                if diff:
+                    return False, "missing in previous {!r}".format(diff)
+            else:
+                return False, "{!r}!={!r}".format(self_deps, other_deps)
+        return True, None
+
     def matches(self, other, deps_matches_subset=False):
-        return (
-            other
-            and self.md5 == other.md5
-            and self.python == other.python
-            and self.version == other.version
-            and self.sitepackages == other.sitepackages
-            and self.usedevelop == other.usedevelop
-            and self.alwayscopy == other.alwayscopy
-            and (
-                all(d in self.deps for d in other.deps)
-                if deps_matches_subset is True
-                else self.deps == other.deps
-            )
-        )
+        outcome, _ = self.matches_with_reason(other, deps_matches_subset)
+        return outcome
 
 
 class VirtualEnv(object):
@@ -163,15 +196,19 @@ class VirtualEnv(object):
             if status string is empty, all is ok.
         """
         rconfig = CreationConfig.readconfig(self.path_config)
-        if (
-            not self.envconfig.recreate
-            and rconfig
-            and rconfig.matches(
-                self._getliveconfig(), getattr(self.envconfig, "deps_matches_subset", False)
-            )
-        ):
+        if self.envconfig.recreate:
+            reason = "-r flag"
+        else:
+            if rconfig is None:
+                reason = "no previous config {}".format(self.path_config)
+            else:
+                live_config = self._getliveconfig()
+                deps_subset_match = getattr(self.envconfig, "deps_matches_subset", False)
+                outcome, reason = rconfig.matches_with_reason(live_config, deps_subset_match)
+        if reason is None:
             action.info("reusing", self.envconfig.envdir)
             return
+        action.info("cannot reuse", reason)
         if rconfig is None:
             action.setactivity("create", self.envconfig.envdir)
         else:
@@ -179,26 +216,33 @@ class VirtualEnv(object):
         try:
             self.hook.tox_testenv_create(action=action, venv=self)
             self.just_created = True
-        except tox.exception.UnsupportedInterpreter:
-            return sys.exc_info()[1]
+        except tox.exception.UnsupportedInterpreter as exception:
+            return exception
         try:
             self.hook.tox_testenv_install_deps(action=action, venv=self)
         except tox.exception.InvocationError as exception:
             return "could not install deps {}; v = {!r}".format(self.envconfig.deps, exception)
 
     def _getliveconfig(self):
-        python = self.envconfig.python_info.executable
-        md5 = getdigest(python)
+        base_resolved_python_path = self.envconfig.python_info.executable
         version = tox.__version__
         sitepackages = self.envconfig.sitepackages
         develop = self.envconfig.usedevelop
         alwayscopy = self.envconfig.alwayscopy
         deps = []
         for dep in self.get_resolved_dependencies():
-            raw_dep = dep.name
-            md5 = getdigest(raw_dep)
-            deps.append((md5, raw_dep))
-        return CreationConfig(md5, python, version, sitepackages, develop, deps, alwayscopy)
+            dep_name_md5 = getdigest(dep.name)
+            deps.append((dep_name_md5, dep.name))
+        base_resolved_python_md5 = getdigest(base_resolved_python_path)
+        return CreationConfig(
+            base_resolved_python_md5,
+            base_resolved_python_path,
+            version,
+            sitepackages,
+            develop,
+            deps,
+            alwayscopy,
+        )
 
     def _getresolvedeps(self):
         warnings.warn(
@@ -225,22 +269,36 @@ class VirtualEnv(object):
         return re.match(self.envconfig.platform, sys.platform)
 
     def finish(self):
-        self._getliveconfig().writeconfig(self.path_config)
+        previous_config = CreationConfig.readconfig(self.path_config)
+        live_config = self._getliveconfig()
+        if previous_config is None or not previous_config.matches(live_config):
+            content = live_config.writeconfig(self.path_config)
+            self.session.report.verbosity1(
+                "write config to {} as {!r}".format(self.path_config, content)
+            )
 
     def _needs_reinstall(self, setupdir, action):
         setup_py = setupdir.join("setup.py")
         setup_cfg = setupdir.join("setup.cfg")
         args = [self.envconfig.envpython, str(setup_py), "--name"]
-        env = self._getenv()
-        output = action.popen(args, cwd=setupdir, redirect=False, returnout=True, env=env)
-        name = output.strip()
-        args = [self.envconfig.envpython, "-c", "import sys; print(sys.path)"]
+        env = self._get_os_environ()
+        output = action.popen(
+            args, cwd=setupdir, redirect=False, returnout=True, env=env, capture_err=False
+        )
+        name = next(
+            (i for i in output.split("\n") if i and not i.startswith("pydev debugger:")), ""
+        )
+        args = [
+            self.envconfig.envpython,
+            "-c",
+            "import sys;  import json; print(json.dumps(sys.path))",
+        ]
         out = action.popen(args, redirect=False, returnout=True, env=env)
         try:
-            sys_path = ast.literal_eval(out.strip())
-        except SyntaxError:
+            sys_path = json.loads(out)
+        except ValueError:
             sys_path = []
-        egg_info_fname = ".".join((name, "egg-info"))
+        egg_info_fname = ".".join((to_filename(name), "egg-info"))
         for d in reversed(sys_path):
             egg_info = py.path.local(d).join(egg_info_fname)
             if egg_info.check():
@@ -261,38 +319,32 @@ class VirtualEnv(object):
 
         return needs_reinstall
 
-    def developpkg(self, setupdir, action):
+    def install_pkg(self, dir, action, name, is_develop=False):
         assert action is not None
+
         if getattr(self, "just_created", False):
-            action.setactivity("develop-inst", setupdir)
+            action.setactivity(name, dir)
             self.finish()
-            extraopts = []
+            pip_flags = ["--exists-action", "w"]
         else:
-            if not self._needs_reinstall(setupdir, action):
-                action.setactivity("develop-inst-noop", setupdir)
+            if is_develop and not self._needs_reinstall(dir, action):
+                action.setactivity("{}-noop".format(name), dir)
                 return
-            action.setactivity("develop-inst-nodeps", setupdir)
-            extraopts = ["--no-deps"]
-
+            action.setactivity("{}-nodeps".format(name), dir)
+            pip_flags = ["--no-deps"] + ([] if is_develop else ["-U"])
+        pip_flags.extend(["-v"] * min(3, action.report.verbosity - 2))
         if action.venv.envconfig.extras:
-            setupdir += "[{}]".format(",".join(action.venv.envconfig.extras))
+            dir += "[{}]".format(",".join(action.venv.envconfig.extras))
+        target = [dir]
+        if is_develop:
+            target.insert(0, "-e")
+        self._install(target, extraopts=pip_flags, action=action)
 
-        self._install(["-e", setupdir], extraopts=extraopts, action=action)
+    def developpkg(self, setupdir, action):
+        self.install_pkg(setupdir, action, "develop-inst", is_develop=True)
 
     def installpkg(self, sdistpath, action):
-        assert action is not None
-        if getattr(self, "just_created", False):
-            action.setactivity("inst", sdistpath)
-            self.finish()
-            extraopts = []
-        else:
-            action.setactivity("inst-nodeps", sdistpath)
-            extraopts = ["-U", "--no-deps"]
-
-        if action.venv.envconfig.extras:
-            sdistpath += "[{}]".format(",".join(action.venv.envconfig.extras))
-
-        self._install([sdistpath], extraopts=extraopts, action=action)
+        self.install_pkg(sdistpath, action, "inst")
 
     def _installopts(self, indexserver):
         options = []
@@ -303,16 +355,36 @@ class VirtualEnv(object):
         return options
 
     def run_install_command(self, packages, action, options=()):
-        argv = self.envconfig.install_command[:]
-        i = argv.index("{packages}")
-        argv[i : i + 1] = packages
-        if "{opts}" in argv:
-            i = argv.index("{opts}")
-            argv[i : i + 1] = list(options)
+        def expand(val):
+            # expand an install command
+            if val == "{packages}":
+                for package in packages:
+                    yield package
+            elif val == "{opts}":
+                for opt in options:
+                    yield opt
+            else:
+                yield val
 
-        for x in ("PIP_RESPECT_VIRTUALENV", "PIP_REQUIRE_VIRTUALENV", "__PYVENV_LAUNCHER__"):
-            os.environ.pop(x, None)
+        cmd = list(chain.from_iterable(expand(val) for val in self.envconfig.install_command))
 
+        self.ensure_pip_os_environ_ok()
+
+        old_stdout = sys.stdout
+        sys.stdout = codecs.getwriter("utf8")(sys.stdout)
+        try:
+            self._pcall(
+                cmd,
+                cwd=self.envconfig.config.toxinidir,
+                action=action,
+                redirect=self.session.report.verbosity < 2,
+            )
+        finally:
+            sys.stdout = old_stdout
+
+    def ensure_pip_os_environ_ok(self):
+        for key in ("PIP_RESPECT_VIRTUALENV", "PIP_REQUIRE_VIRTUALENV", "__PYVENV_LAUNCHER__"):
+            os.environ.pop(key, None)
         if "PYTHONPATH" not in self.envconfig.passenv:
             # If PYTHONPATH not explicitly asked for, remove it.
             if "PYTHONPATH" in os.environ:
@@ -322,17 +394,11 @@ class VirtualEnv(object):
                 )
                 os.environ.pop("PYTHONPATH")
 
-        old_stdout = sys.stdout
-        sys.stdout = codecs.getwriter("utf8")(sys.stdout)
-        try:
-            self._pcall(
-                argv,
-                cwd=self.envconfig.config.toxinidir,
-                action=action,
-                redirect=self.session.report.verbosity < 2,
-            )
-        finally:
-            sys.stdout = old_stdout
+        # installing packages at user level may mean we're not installing inside the venv
+        os.environ["PIP_USER"] = "0"
+
+        # installing without dependencies may lead to broken packages
+        os.environ["PIP_NO_DEPS"] = "0"
 
     def _install(self, deps, extraopts=None, action=None):
         if not deps:
@@ -359,13 +425,13 @@ class VirtualEnv(object):
                 options.extend(extraopts)
             self.run_install_command(packages=packages, options=options, action=action)
 
-    def _getenv(self, testcommand=False):
-        if testcommand:
+    def _get_os_environ(self, is_test_command=False):
+        if is_test_command:
             # for executing tests we construct a clean environment
             env = {}
-            for envname in self.envconfig.passenv:
-                if envname in os.environ:
-                    env[envname] = os.environ[envname]
+            for env_key in self.envconfig.passenv:
+                if env_key in os.environ:
+                    env[env_key] = os.environ[env_key]
         else:
             # for executing non-test commands we use the full
             # invocation environment
@@ -377,22 +443,34 @@ class VirtualEnv(object):
         env["VIRTUAL_ENV"] = str(self.path)
         return env
 
-    def test(self, redirect=False):
-        with self.session.newaction(self, "runtests") as action:
-            self.status = 0
-            self.session.make_emptydir(self.envconfig.envtmpdir)
-            self.envconfig.envtmpdir.ensure(dir=1)
+    def test(
+        self,
+        redirect=False,
+        name="runtests",
+        commands=None,
+        ignore_outcome=None,
+        ignore_errors=None,
+        display_hash_seed=False,
+    ):
+        if commands is None:
+            commands = self.envconfig.commands
+        if ignore_outcome is None:
+            ignore_outcome = self.envconfig.ignore_outcome
+        if ignore_errors is None:
+            ignore_errors = self.envconfig.ignore_errors
+        with self.session.newaction(self, name) as action:
             cwd = self.envconfig.changedir
-            env = self._getenv(testcommand=True)
-            # Display PYTHONHASHSEED to assist with reproducibility.
-            action.setactivity("runtests", "PYTHONHASHSEED={!r}".format(env.get("PYTHONHASHSEED")))
-            for i, argv in enumerate(self.envconfig.commands):
+            if display_hash_seed:
+                env = self._get_os_environ(is_test_command=True)
+                # Display PYTHONHASHSEED to assist with reproducibility.
+                action.setactivity(name, "PYTHONHASHSEED={!r}".format(env.get("PYTHONHASHSEED")))
+            for i, argv in enumerate(commands):
                 # have to make strings as _pcall changes argv[0] to a local()
                 # happens if the same environment is invoked twice
                 message = "commands[{}] | {}".format(
                     i, " ".join([pipes.quote(str(x)) for x in argv])
                 )
-                action.setactivity("runtests", message)
+                action.setactivity(name, message)
                 # check to see if we need to ignore the return code
                 # if so, we need to alter the command line arguments
                 if argv[0].startswith("-"):
@@ -411,10 +489,10 @@ class VirtualEnv(object):
                         action=action,
                         redirect=redirect,
                         ignore_ret=ignore_ret,
-                        testcommand=True,
+                        is_test_command=True,
                     )
                 except tox.exception.InvocationError as err:
-                    if self.envconfig.ignore_outcome:
+                    if ignore_outcome:
                         msg = "command failed but result from testenv is ignored\ncmd:"
                         self.session.report.warning("{} {}".format(msg, err))
                         self.status = "ignored failed command"
@@ -422,7 +500,7 @@ class VirtualEnv(object):
 
                     self.session.report.error(str(err))
                     self.status = "commands failed"
-                    if not self.envconfig.ignore_errors:
+                    if not ignore_errors:
                         break  # Don't process remaining commands
                 except KeyboardInterrupt:
                     self.status = "keyboardinterrupt"
@@ -430,19 +508,32 @@ class VirtualEnv(object):
                     raise
 
     def _pcall(
-        self, args, cwd, venv=True, testcommand=False, action=None, redirect=True, ignore_ret=False
+        self,
+        args,
+        cwd,
+        venv=True,
+        is_test_command=False,
+        action=None,
+        redirect=True,
+        ignore_ret=False,
+        returnout=False,
     ):
+        # construct environment variables
         os.environ.pop("VIRTUALENV_PYTHON", None)
+        env = self._get_os_environ(is_test_command=is_test_command)
+        bin_dir = str(self.envconfig.envbindir)
+        env["PATH"] = os.pathsep.join([bin_dir, os.environ["PATH"]])
+        self.session.report.verbosity2("setting PATH={}".format(env["PATH"]))
 
-        cwd.ensure(dir=1)
+        # get command
         args[0] = self.getcommandpath(args[0], venv, cwd)
         if sys.platform != "win32" and "TOX_LIMITED_SHEBANG" in os.environ:
             args = prepend_shebang_interpreter(args)
-        env = self._getenv(testcommand=testcommand)
-        bindir = str(self.envconfig.envbindir)
-        env["PATH"] = p = os.pathsep.join([bindir, os.environ["PATH"]])
-        self.session.report.verbosity2("setting PATH={}".format(p))
-        return action.popen(args, cwd=cwd, env=env, redirect=redirect, ignore_ret=ignore_ret)
+
+        cwd.ensure(dir=1)  # ensure the cwd exists
+        return action.popen(
+            args, cwd=cwd, env=env, redirect=redirect, ignore_ret=ignore_ret, returnout=returnout
+        )
 
 
 def getdigest(path):
@@ -510,6 +601,32 @@ def tox_runtest(venv, redirect):
     return True  # Return non-None to indicate plugin has completed
 
 
+@tox.hookimpl
+def tox_runtest_pre(venv):
+    venv.status = 0
+    venv.session.make_emptydir(venv.envconfig.envtmpdir)
+    venv.envconfig.envtmpdir.ensure(dir=1)
+    venv.test(
+        name="run-test-pre",
+        commands=venv.envconfig.commands_pre,
+        redirect=False,
+        ignore_outcome=False,
+        ignore_errors=False,
+        display_hash_seed=True,
+    )
+
+
+@tox.hookimpl
+def tox_runtest_post(venv):
+    venv.test(
+        name="run-test-post",
+        commands=venv.envconfig.commands_post,
+        redirect=False,
+        ignore_outcome=False,
+        ignore_errors=False,
+    )
+
+
 @tox.hookimpl
 def tox_runenvreport(venv, action):
     # write out version dependency information
